#!/bin/sh

#;=========================================================================
#; This is free and unencumbered software released into the public domain.
#;
#; Anyone is free to copy, modify, publish, use, compile, sell, or
#; distribute this software, either in source code form or as a compiled
#; binary, for any purpose, commercial or non-commercial, and by any
#; means.
#;
#; In jurisdictions that recognize copyright laws, the author or authors
#; of this software dedicate any and all copyright interest in the
#; software to the public domain. We make this dedication for the benefit
#; of the public at large and to the detriment of our heirs and
#; successors. We intend this dedication to be an overt act of
#; relinquishment in perpetuity of all present and future rights to this
#; software under copyright law.
#;
#; THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
#; EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
#; MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT.
#; IN NO EVENT SHALL THE AUTHORS BE LIABLE FOR ANY CLAIM, DAMAGES OR
#; OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE,
#; ARISING FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR
#; OTHER DEALINGS IN THE SOFTWARE.
#;
#; For more information, please refer to <http://unlicense.org/>
#;=========================================================================

#;-------------------------------------------------------------------------
#; 2014-01-07	initial source code release by Tinybit <tinybit@tom.com>
#;-------------------------------------------------------------------------

if [ "$1" = "" ]; then
	# write error message to /dev/stderr
	echo "Please use \`make' to build images. exit!" 1>&2
	exit 1
elif [ "$1" = "mbr_8G.img" ]; then
	# 0x3F00=16128
	fatsize=16128
elif [ "$1" = "mbr_16G.img" ]; then
	# 0x7E00=32256
	fatsize=32256
elif [ "$1" = "mbr_32G.img" ]; then
	# 0xFC00=64512
	fatsize=64512
elif [ "$1" = "mbr_64G.img" ]; then
	# 0x1F800=129024
	fatsize=129024
elif [ "$1" = "mbr_128G.img" ]; then
	# 0x3F000=258048
	fatsize=258048
else
	# write error message to /dev/stderr
	echo "Invalid image file name: $1, exit!" 1>&2
	exit 1
fi

if [ ! -f multimbr.mbr ]; then
	# write error message to /dev/stderr
	echo "Cannot find multimbr.mbr, exit!" 1>&2
	exit 1
fi

if [ ! -f grldr ]; then
	# write error message to /dev/stderr
	echo "Cannot find grldr, exit!" 1>&2
	exit 1
fi

if [ ! -f preset_menu.lst ]; then
	# write error message to /dev/stderr
	echo "Cannot find preset_menu.lst, exit!" 1>&2
	exit 1
fi

# locate bootlace signature
i=$((0x2000+$(od -A n -j 0x206C -N 4 -t u4 grldr)-0x8200))
j=$((0xce1a02b0-$(od -A n -j $i -N 4 -t u4 grldr)))

if [ "$j" != 0 ]; then
	# write error message to /dev/stderr
	echo "Unrecognized grldr, exit!" 1>&2
	exit 1
fi

if [ -f grldr.tmp ]; then
	if ! rm grldr.tmp; then
		# write error message to /dev/stderr
		echo "Could not remove grldr.tmp, exit!" 1>&2
		exit 1
	fi
fi

if [ ! -f grldr.tmp ] || [ grldr.tmp -ot grldr ]; then
	# write message to /dev/stderr
	# echo Updating grldr.tmp ... 1>&2

	{
		# duplicate the head
		dd bs=8197 skip=0 count=1 2> /dev/null

		# discard the control byte
		dd bs=1 count=1 2> /dev/null 1>&2

		# write new control byte and disable PXE
		printf "\x01"

		# duplicate the body and locate the preset menu
		dd bs=$((i+16-8198)) count=1 2> /dev/null

		# write our new preset menu.
		cat preset_menu.lst

		# end the menu with a null
		printf "\x00"

	} < grldr > grldr.tmp
fi

# Some tests about exec and IO redirection
# echo test begin
# echo "exec > $1"
# exec > $1
# echo test123 end
# exec > another
# echo another test
# Output to the stdin and restore the original stdout
# exec >&0
# echo back to stdout
# echo ok!
# Confirm stdin is still normal ...
# cat -
# echo input
# exit

# Redirect stdout to destination image file
exec > $1

# two FATs, data area
# data area=fatsize * clusters_per_FAT32_sector * sectors_per_cluster
# clusters_per_FAT32_sector = 512 / 4 = 128
# sectors_per_cluster = 8
# 128 * 8 = 1024
# data area = fatsize * 1024
# two FATs = fatsize * 2
# two FATs + data area = fatsize * 1026
t=$((fatsize*1026))

i=16224; #; i=$((33+126+255*63))
j=0
a=0
b=0

cat grldr.tmp /dev/zero | while [ $i -gt 31 ]; do
	a=$((i&255))
	b=$(((i>>8)&255))
	k=$((i+t))
	l=$((k-63))
	c=$((k&255))
	d=$(((k>>8)&255))
	p=$(((k>>16)&255))
	q=$(((k>>24)&255))
	y=$((l&255))
	z=$(((l>>8)&255))
	r=$(((l>>16)&255))
	s=$(((l>>24)&255))
	m=$((i-32+1))
	n=$((i-32+6))
	e=$((m&255))
	f=$(((m>>8)&255))
	g=$((n&255))
	h=$(((n>>8)&255))

    if [ $j -lt 190 ]; then
	#; 190 MBR sectors (stage 1)
	printf "\xEB\x5E\x90\x4D\x53\x57\x49\x4E\x34\x2E\x31\x00\x02\x08\x$(printf %x $a)\x$(printf %x $b)\x02\x00\x00\x00\x00\xF8\x00\x00\x3F\x00\xFF\x00\x3F\x00\x00\x00\x$(printf %x $c)\x$(printf %x $d)\x$(printf %x $p)\x$(printf %x $q)\x$(printf %x $((fatsize&255)))\x$(printf %x $(((fatsize>>8)&255)))\x$(printf %x $(((fatsize>>16)&255)))\x$(printf %x $(((fatsize>>24)&255)))\x00\x00\x00\x00\x02\x00\x00\x00\x$(printf %x $e)\x$(printf %x $f)\x$(printf %x $g)\x$(printf %x $h)"
	dd if=multimbr.mbr bs=1 skip=52 count=406 2> /dev/null
	printf "\x$(printf %x $y)\x$(printf %x $z)\x$(printf %x $r)\x$(printf %x $s)"

	if true; then
		# this will not create the second partition
		printf "\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00"
	else
		# this will create the second partition
		# 0xFFFFFFFF=4294967295
		len=$((4294967295-k))
		printf "\x00\xFE\xFF\xFF\x0C\xFE\xFF\xFF\x$(printf %x $c)\x$(printf %x $d)\x$(printf %x $p)\x$(printf %x $q)\x$(printf %x $((len&255)))\x$(printf %x $(((len>>8)&255)))\x$(printf %x $(((len>>16)&255)))\x$(printf %x $(((len>>24)&255)))"

	fi
	printf "\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x55\xAA"
	true
    elif [ $j -eq 190 ]; then
	#; 1 sector of stage 2
	dd if=multimbr.mbr skip=1 count=1 2> /dev/null
	#; 1151 sectors for GRLDR(stage 3). It begins at sector 191.
	dd bs=14 count=1 2> /dev/null
	true
    elif [ $j -lt 1341 ]; then
	printf "\x$(printf %x $a)\x$(printf %x $b)"
	dd bs=510 count=1 2> /dev/null
	true
    elif [ $j -eq 1341 ]; then
	printf "\x$(printf %x $a)\x$(printf %x $b)"
	dd bs=496 count=1 2> /dev/null
	#; end of GRLDR at sector 1341
	true
    elif [ -f userdata ]; then
	break
	true
    else
	#; 14851 empty sectors, from sector 1342 to sector 16192
	printf "\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x$(printf %x $a)\x$(printf %x $b)"
	dd if=/dev/zero bs=496 count=1 2> /dev/null
	true
    fi

	i=$((i-1))
	j=$((j+1))
done

#; debug output: i=16224, j=0, a=0, b=0
#; echo i=$i j=$j a=$a b=$b 1>&2
#; exit

if [ -f userdata ]; then
	i=14882; #; i=$((33+126+255*63-1342))
	j=1342
    cat userdata /dev/zero | while [ $i -gt 31 ]; do
	a=$((i&255))
	b=$(((i>>8)&255))

	if [ $j -eq 1342 ]; then
	    #; 14851 sectors for USERDATA, from sector 1342 to sector 16192.
	    dd bs=14 count=1 2> /dev/null
	    printf "\x$(printf %x $a)\x$(printf %x $b)"
	    dd bs=510 count=1 2> /dev/null
	    true
	elif [ $j -lt 16192 ]; then
	    printf "\x$(printf %x $a)\x$(printf %x $b)"
	    dd bs=510 count=1 2> /dev/null
	    true
	elif [ $j -eq 16192 ]; then
	    printf "\x$(printf %x $a)\x$(printf %x $b)"
	    dd bs=496 count=1 2> /dev/null
	    #; end of USERDATA at sector 16192
	    true
	else
	    echo "It should not get here. Exit!" 1>&2
	    exit 1
	    true
	fi

	i=$((i-1))
	j=$((j+1))
    done
fi

#; write 31 hidden sectors
printf "\x52\x52\x61\x41"
dd if=/dev/zero bs=480 count=1 2> /dev/null
printf "\x72\x72\x41\x61\xff\xff\xff\xff\xff\xff\xff\xff\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x55\xAA"
dd if=/dev/zero bs=510 count=1 2> /dev/null
printf "\x55\xAA"
dd if=/dev/zero bs=2046 count=1 2> /dev/null
printf "\x55\xAA\x52\x52\x61\x41"
dd if=/dev/zero bs=480 count=1 2> /dev/null
printf "\x72\x72\x41\x61\xff\xff\xff\xff\xff\xff\xff\xff\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x55\xAA"
dd if=/dev/zero bs=510 count=1 2> /dev/null
printf "\x55\xAA"
dd if=/dev/zero bs=11776 count=1 2> /dev/null

#; write the 1st FAT
printf "\xF8\xFF\xFF\x0F\xFF\xFF\xFF\xFF\xFF\xFF\xFF\x0F"
dd if=/dev/zero bs=$((fatsize*512-12)) count=1 2> /dev/null

#; write the 2nd FAT, and 8 more sectors(=1 cluster) for root dir
printf "\xF8\xFF\xFF\x0F\xFF\xFF\xFF\xFF\xFF\xFF\xFF\x0F"
dd if=/dev/zero bs=$((fatsize*512-12+4096)) count=1 2> /dev/null

