#;=========================================================================
#; This is free and unencumbered software released into the public domain.
#;
#; Anyone is free to copy, modify, publish, use, compile, sell, or
#; distribute this software, either in source code form or as a compiled
#; binary, for any purpose, commercial or non-commercial, and by any
#; means.
#;
#; In jurisdictions that recognize copyright laws, the author or authors
#; of this software dedicate any and all copyright interest in the
#; software to the public domain. We make this dedication for the benefit
#; of the public at large and to the detriment of our heirs and
#; successors. We intend this dedication to be an overt act of
#; relinquishment in perpetuity of all present and future rights to this
#; software under copyright law.
#;
#; THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
#; EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
#; MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT.
#; IN NO EVENT SHALL THE AUTHORS BE LIABLE FOR ANY CLAIM, DAMAGES OR
#; OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE,
#; ARISING FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR
#; OTHER DEALINGS IN THE SOFTWARE.
#;
#; For more information, please refer to <http://unlicense.org/>
#;=========================================================================

#;-------------------------------------------------------------------------
#; 2014-01-07	initial source code release by Tinybit <tinybit@tom.com>
#;-------------------------------------------------------------------------


#undef	SUPPORT_LBA
#define SUPPORT_LBA

	.text

	.code16

start:
	jmp	1f

	. = start + 0x02

	.byte	0x90

	. = start + 0x03

	.ascii	"MSWIN4.1"

	. = start + 0x0B

	.word	0x200

	. = start + 0x0D

	/* Sectors per cluster. Valid values are 1, 2, 4, 8, 16, 32, 64 and 128.
	 * But a cluster size larger than 32K should not occur.
	 */

	.byte	0x08	// sectors per cluster, should be changed by installer

	. = start + 0x0E

	.word	0x0020	// reserved sectors, should be changed by installer

	. = start + 0x10

	.byte	0x02	// number of FATs(nearly always 2)

	. = start + 0x11

	.word	0	// FAT12/16 root dir entries. Must be 0 for FAT32

	. = start + 0x13

	.word	0	// FAT12/16 total sectors. Must be 0 for FAT32

	. = start + 0x15

	.byte	0xF8	// media descriptor

	. = start + 0x16

	.word	0	// FAT12/16 sectors per FAT. Must be 0 for FAT32

	. = start + 0x18

	.word	0x3f	// sectors per track

	. = start + 0x1A

	.word	0xff	// number of heads

	. = start + 0x1C

	.long	0x3F	// hidden sectors, installer will change it

	. = start + 0x20

	.long	0x00FC0000	// total sectors, installer will change it

	. = start + 0x24

	/* 4-byte FAT32 sectors per FAT, installer will change it */

	/* The byte at offset 0x026 in this entry should never become
	 * 0x28 or 0x29 in order to avoid any misinterpretation with the
	 * EBPB format under non-FAT32 aware operating systems.
	 */

	.byte	0x00, 0x3F, 0x00, 0x00
	//.byte	0x80, 0x00, 0x29, 0x00
	//.long	0

	. = start + 0x28

	.word	0	// current active FAT, installer will change it

	. = start + 0x2A

	.word	0	// FAT32 revision number, installer will change it

	. = start + 0x2C

	.long	2	// root dir starting cluster, installer will change it

	. = start + 0x30

	.word	1	// file system information sector number
			// installer will change it

	. = start + 0x32

	.word	6	// boot record backup sector number
			// installer will change it

	. = start + 0x34

	.long	0	// reserved
	.long	0	// reserved
	.long	0	// reserved

	. = start + 0x40

	.byte	0x80	// drive number

	. = start + 0x41

	/* Windows NT use it for CHKDSK flags:
	 * Bits 7-2 always cleared;
	 * Bit 1: disk I/O errors encountered, possible bad sectors,
		  run surface scan on next boot;
	 * Bit 0: volume is "dirty" and was not properly unmounted before
		  shutdown, run CHKDSK on next boot.
	 */

	.byte	0	// reserved, should be set to 0 by formating tools.

	. = start + 0x42

	/* Signature (must be 28h or 29h to be recognised by NT). */

	.byte	0x29		// extended boot signature for FAT32

	. = start + 0x43

	/* Originally it was 4-byte volume serial number. */

	/* hi word at offset 0x45 is used for serial number of sector 190 */

	.word	0xAF63		// hour, minutes, seconds

	.word	(0x21 + 126 + (255*63) - 190)	// installer will change it

	. = start + 0x47

	.ascii	"NO NAME    "	// volume label, 11 bytes

	. = start + 0x52

	.ascii	"FAT32   "	// filesystem ID, 8 bytes

	. = start + 0x5A

	.word	0	// reserved
	.word	0	// reserved
	.word	0	// reserved

	. = start + 0x60

1:
	xorw	%ax, %ax
	movw	$0x7C00, %bp	/* constant BP=0x7C00 */

	movw	%ax, %ss	/* constant SS=0 */
	movw	%bp, %sp

	/* save DL first */
	movb	%dl, 0x40(%bp)	/* assume BIOS pass correct drive number. */

	/* Got control! */

	/* Must switch to 80x25 color text mode! */
	pushaw
	movb	$0x03, %al	/* AX=0x0003 */
	int	$0x10
	popaw

	movw	%ax, %ds	/* constant DS=0 */
	xchgw	%ax, %di	/* DI=0 */

	/* print DX */
	xchgw	%ax, %dx	/* store DX into AX */
	call	print_hex	# BX changed

	/* print serial number of this sector */
	movw	0x0E(%bp), %ax
	call	print_hex	# BX changed

	/* sleep 3 seconds */

	sti			/* enable interrupt for int8 timer */
	movw	$55, %cx	/* 55 / 18.2 = 3 seconds */
2:
	movw	0x46c, %ax	/* get new ticks */
	cmpw	%ax, %bx	/* changed? */
	je	2b		/* no, continue to wait */
	/* ticks changed by int8 */
	xchgw	%ax, %bx	/* update ticks */
	loop	2b

	/* CX=0 */

	pushw	$0x0201	//---	/* BP-2 point to int13/AX: read 1 sector */

	/* read C/H/S=0/0/1 to 0000:7E00 */
	/* Generally, it is safe enough. */
	xorw	%ax, %ax
	movw	%ax, %es
	movw	$0x7E00, %bx
	pushw	%es	//---
	pushw	%bx	//---	/* BP-6 point to 0x00007E00 on stack */
//2:
//	lesw	-6(%bp), %bx	/* ES:BX=0000:7E00 */
	incw	%cx		/* read 1 sector */
	call	readDisk	/* CX=0, DX=0 */

//	pushw	%ax
//	movw	0x0E(%bx), %ax	# test serial number
//	call	print_hex	# BX changed
//	popw	%ax
//	cmpw	$128, %ax
//	jb	2b

//	movw	$0xEEEE, %ax	# test end
//	jmp	boot_error

	//pushw	%cx		/* 1, no POP in LBA branch. See below. */
	//incw	%cx		/* CX=2 */
	//movw	%cx, 0x18(%bp)	/* interim sectors per track */
	//movw	%cx, 0x1a(%bp)	/* interim number of heads */
	//pushw	%cx		/* 2, no POP in LBA branch. See below. */

	movw	0x0E(%bx), %si	# mbr0

	/* print serial number of mbr0 */
	movw	%si, %ax
	call	print_hex	# BX changed

#ifdef	SUPPORT_LBA
	/* check LBA support */
	/* This can also be considered safe. */
       
	movb	$0x41, %ah
	movw	$0x55AA, %bx
	call	int13
	jc	1f		/* no ebios */
	cmpw	$0xAA55, %bx
	jnz	1f		/* no ebios */
	//testb	$1, %cl
	//jz	1f
	.byte	0xD0, 0xD9	# RCR CL,1; D0 C9  ROR CL,1 also OK
	jnc	1f		/* no ebios */

#if 0
	movb	$0x42, -1(%bp)	/* LBA supported */
#else
	/* Use CHS mode to load GRLDR by default. */
#endif

	/* print "AA55", indicating LBA support. */
	xchgw	%ax, %bx	/* store BX into AX */
	call	print_hex	# BX changed

#if 0
	/* Keep the pushed 1 and 2 on stack without POPs. There is no harm. */
	jmp	geometry_OK	/* ready to load GRLDR */
#else
	/* pass through and continue to check CHS mode. */
#endif

1:
#endif

	/* CHS mode. Must determine geometry. */

	/* read C/H/S=0/1/1 to 0000:7E00 */
	//movw	$2, %ax		/* AX=2, interim sectors per track */
	//popw	%ax		/* AX=2, interim sectors per track */
	movw	0x18(%bp), %ax	/* AX=2, interim sectors per track */
	lesw	-6(%bp), %bx	/* ES:BX=0000:7E00 */
	//popw	%cx		/* CX=1, read 1 sector */
	incw	%cx		/* read 1 sector */
	call	readDisk	# CX=0, DX=0

	movw	%si, %ax	# mbr0
	subw	0x0E(%bx), %ax	/* real sectors per track */
	jbe	boot_error	/* installation problem? */
	cmpw	$63, %ax
	ja	boot_error	/* installation problem? */

	movw	%ax, 0x18(%bp)	/* update sectors per track! */

	/* print real sectors per track */
	call	print_hex	# BX changed

	/* read C/H/S=1/0/1 to 0000:7E00 */
	//shlw	$1, %ax		/* AX=interim sectors per cylinder */
	mulw	0x1A(%bp)	/* AX=interim sectors per cylinder */
				/* DX=0 */
	lesw	-6(%bp), %bx	/* ES:BX=0000:7E00 */
	//movw	$1, %cx		/* read 1 sector */
	incw	%cx		/* read 1 sector */
	call	readDisk	# CX=0, DX=0

	movw	%si, %ax	# mbr0
	subw	0x0E(%bx), %ax	/* real sectors per cylinder */
	jbe	boot_error	/* installation problem? */

	/* DX=0 */
	divw	0x18(%bp)	/* rem=DX, should be 0 */
				/* quo=AX, number of heads */

	/* if DX != 0, stop! */
	testw	%dx, %dx
	jnz	boot_error	/* installation problem? */

	movw	%ax, 0x1A(%bp)	/* update number of heads! */

	/* print real number of heads */
	call	print_hex	# BX changed

	testb	%ah, %ah
	jnz	boot_error	/* installation problem? */

geometry_OK:

	/* read GRLDR to 0000:7E00 */
	/* calculate grldr starting sector number */
	/* grldr serial number = (0x21 + 126 + 255*63 - 190) */
	xchgw	%ax, %si	# AX=mbr0
#if 0
	movw	$(0x21 + 126 + (255*63) - 190), %si
#else
	movw	0x45(%bp), %si
#endif
	subw	%si, %ax
	jbe	boot_error	/* installation problem? */
	lesw	-6(%bp), %bx	/* ES:BX=0000:7E00 */
	movw	$1089, %cx	/* number of sectors for stage2 and grldr */
	call	readDisk	/* CX=0, DX=0 */

	cmpw	%si, 0x0E(%bx)
	jne	boot_error	/* installation problem? */
#if 0
	/* We can do it later in the second stage. */
	movb	0x40(%bp), %dl
	movb	$0xFF, %dh
	pushw	%dx
#endif

#if 0
	movw	$0xACED, %ax	/* OK! passed. */
	jmp	boot_error	/* Test succeeded. Halt. */
#else
	/* jump to stage 2 */
	//ljmp	$0, $0x7E00
	pushw	%ds		/* DS=0 */
	pushw	%bx		/* BX=0x7E00 */
	lret
#endif

int13_retry:
	/* first, try 4 times for CHS mode. if still failed then continue
	 * to try 4 times for LBA, even if no LBA support! */
	pushw	%di
	movw	$8, %di		/* try 8 times on failure */
2:
	pushaw
	cmpw	$4, %di
	ja	1f
	xorb	$0x40, %ah	/* toggle between normal BIOS and EBIOS */
1:
	call	int13
	popaw
	jc	1f
	popw	%di
	ret
1:
	pushaw
	cbw			/* AH=0, reset disk */
	call	int13
	popaw
	decw	%di
	jnz	2b
	popw	%di

disk_error:
	//movw	$0xDEAD, %ax
	movw	8(%si), %ax
	call	print_hex	# BX changed
	movw	$0xCEDE, %ax	# disk read error, give up
boot_error:
	call	print_hex	# BX changed
2:
	jmp	2b

/* Read sectors from disk, using LBA or CHS
 * input:	AX    = 16-bit LBA sector number
 *		CX    = number of sectors to read
 *		AX + CX <= 0x8000
 *		ES:BX = destination buffer
 *		SS    = DS
 *		BP-0x02 -> function call number for AX
 *		BP+0x18 -> sectors per track
 *		BP+0x1A -> number of heads
 *		BP+0x40 -> drive number
 *
 * output:	No return on error
 *		BX not changed
 *		ES    = ES + 0x20 * CX
 *		AX    = AX + CX
 *		CX    = 0
 *		DX    = 0
 *		ZF    = 0
 *		CF    = 0
 */
 
readDisk:
	cwd			/* DX=0 */
2:
	pushaw
#ifdef	SUPPORT_LBA
	pushw	%dx		/* 0 */
	pushw	%dx		/* 0 */
	pushw	%dx		/* 0 */
	pushw	%ax		/* lo 16bit of sector number */
	pushw	%es		/* buffer segment */
	pushw	%bx		/* buffer offset */
	pushw	$1		/* 1 sector to read */
	pushw	$16		/* size of this parameter block */

	cmpb	$0x42, -1(%bp)	/* LBA supported */
	je	1f		/* LBA mode */
#endif

	/* CHS mode */
	divw	0x18(%bp)	/* rem=DX */
				/* quo=AX */
				/* DH=0 */
	incw	%dx		/* DL=sector number */
	xchgw	%dx, %cx	/* CH=0, CL=sector number */

	cwd			/* DX=0 */
	divw	0x1A(%bp)	/* rem=DX */
				/* quo=AX, cylinder number */
				/* DH=0, DL=head number */

	xchgb	%dl, %dh	/* DL=0, DH=head number */
	xchgb	%al, %ch	/* AL=0, CH=lo 8bit cylinder */
#if 1
	shrw	$2, %ax		/* hi 2bit cylinder ... */
	orb	%al, %cl	/* ... should be in CL */
	testb	%ah, %ah
#else
	/* equivalently, this should also work. */
	rolw	$6, %ax		/* hi 2bit cylinder in AH */
	orb	%ah, %cl	/* put into CL */
	testb	%al, %al
#endif
	jnz	boot_error	/* Cylinder overflow! Stop misreading! */
1:
	movw	-2(%bp), %ax	/* (E)BIOS read 1 sector */
#ifdef	SUPPORT_LBA
	movw	%sp, %si	/* DS:SI points to disk address packet */
#endif
	//movb	0x40(%bp), %dl	/* drive number */
	pushw	%es
	call	int13_retry
	popw	%bx
	//jc	disk_error
	leaw	0x20(%bx), %bx
	movw	%bx, %es
#ifdef	SUPPORT_LBA
	popaw			/* remove parameter block from stack */
#endif
	popaw
	incw 	%ax		/* next sector, here ZF=0 */
	loop	2b
	ret

//	. = . - (. - readDisk)/68


int13:
	pushw	%ds
	pushw	%es
//	pushw	%bx	/* int13/ah=41h use BX, so do not save. */
//	pushw	%dx	/* drive number already saved in memory. */
	pushw	%si
	pushw	%di
	pushw	%bp
	stc
	movb	0x40(%bp), %dl	/* drive number */
	int	$0x13
	sti
	popw	%bp
	popw	%di
	popw	%si
//	popw	%dx
//	popw	%bx
	popw	%es
	popw	%ds
	ret

print_hex:
	# input		AX = value to print
	#		DI = video text offset
	#
	# output	ES = 0xB800 + (12*160 + 32)/16
	#		DF = 0
	#		DI = changed
	#		BX = changed
	pushw	$(0xB800 + (12*160 + 32)/16)
	popw	%es		# text mode video memory

	cld
	pushw	%ax
	pushw	%cx
	pushw	%bx
	movw	$4, %cx		# will print 4 hex digits
	xchgw	%ax, %bx	# save AX to BX
2:
	rolw	$4, %bx
#if 0
	movw	%bx, %ax
	andb	$0x0F, %al
	cbw			# AH=0
	cmpb	$10, %al
	jb	1f
	addb	$7, %al
1:
	addw	$0x1F30, %ax	# back/fore ground=blue/bright white
#else
	/* this can save 2 bytes of code. */
	movzbw	%bl, %ax	# AH=0
	andb	$0x0F, %al	# need this. Too bad.
	sahf			# AF=0
	aaa			# AAA operation:
				// { IF (((AL & 0x0F) > 9) || AF == 1)
				//	{ AL += 6; AH++ ; AF = CF = 1; }
				//   ELSE
				//	{ AF = CF = 0; }
				//   AL &= 0x0F; }
	add	%ah, %al
	.byte	0xD5, 0x10	# AAD imm8 operation:
				// { AL += (AH * imm8); AH = 0; }
	addw	$0x1F30, %ax	# back/fore ground=blue/bright white
#endif
	stosw
	loop	2b
	movb	$0x20, %al	# keep colors in AH untouched
	stosw
	popw	%bx
	popw	%cx
	popw	%ax
	ret

	. = start + 0x1B8

	. = start + 0x1BE

	.byte	0x80, 1, 1, 0, 0x0C, 0xFE, 0xFF, 0xFF
	.byte	0x3F, 0, 0, 0, 0x00, 0x00, 0xFC, 0x00

	. = start + 0x1FE

	.word	0xAA55

	/* Stage2 */

	. = start + 0x200

	/* We are loaded at 0000:7E00 */

	jmp	1f

	. = start + 0x202

	.byte	0x90

	. = start + 0x203

	.ascii	"MSWIN4.1"

	. = start + 0x20B

	.word	0x200

	. = start + 0x20D

	/* Sectors per cluster. Valid values are 1, 2, 4, 8, 16, 32, 64 and 128.
	 * But a cluster size larger than 32K should not occur.
	 */

	.byte	0x08	// sectors per cluster, should be changed by installer

	. = start + 0x20E

	.word	(0x21 + 126 + (255*63) - 190)	// reserved sectors

	. = start + 0x210

	.byte	0x02	// number of FATs(nearly always 2)

	. = start + 0x211

	.word	0	// FAT12/16 root dir entries. Must be 0 for FAT32

	. = start + 0x213

	.word	0	// FAT12/16 total sectors. Must be 0 for FAT32

	. = start + 0x215

	.byte	0xF8	// media descriptor

	. = start + 0x216

	.word	0	// FAT12/16 sectors per FAT. Must be 0 for FAT32

	. = start + 0x218

	.word	0x3f	// sectors per track

	. = start + 0x21A

	.word	0xff	// number of heads

	. = start + 0x21C

	.long	0x3F	// hidden sectors, installer will change it

	. = start + 0x220

	.long	0x00FC0000	// total sectors, installer will change it

	. = start + 0x224

	/* 4-byte FAT32 sectors per FAT, installer will change it */

	/* The byte at offset 0x026 in this entry should never become
	 * 0x28 or 0x29 in order to avoid any misinterpretation with the
	 * EBPB format under non-FAT32 aware operating systems.
	 */

	.byte	0x00, 0x3F, 0x00, 0x00
	//.byte	0x80, 0x00, 0x29, 0x00
	//.long	0

	. = start + 0x228

	.word	0	// current active FAT, installer will change it

	. = start + 0x22A

	.word	0	// FAT32 revision number, installer will change it

	. = start + 0x22C

	.long	2	// root dir starting cluster, installer will change it

	. = start + 0x230

	.word	1	// file system information sector number
			// installer will change it

	. = start + 0x232

	.word	6	// boot record backup sector number
			// installer will change it

	. = start + 0x234

	.long	0	// reserved
	.long	0	// reserved
	.long	0	// reserved

	. = start + 0x240

	.byte	0x80	// drive number

	. = start + 0x241

	/* Windows NT use it for CHKDSK flags:
	 * Bits 7-2 always cleared;
	 * Bit 1: disk I/O errors encountered, possible bad sectors,
		  run surface scan on next boot;
	 * Bit 0: volume is "dirty" and was not properly unmounted before
		  shutdown, run CHKDSK on next boot.
	 */

	.byte	0	// reserved, should be set to 0 by formating tools.

	. = start + 0x242

	/* Signature (must be 28h or 29h to be recognised by NT). */

	.byte	0x29		// extended boot signature for FAT32

	. = start + 0x243

	/* Originally it was 4-byte volume serial number. */

	/* hi word at offset 0x45 is used for serial number of sector 190 */

	.word	0xAF63		// hour, minutes, seconds

	.word	(0x21 + 126 + (255*63) - 190)	// installer will change it

	. = start + 0x247

	.ascii	"NO NAME    "	// volume label, 11 bytes

	. = start + 0x252

	.ascii	"FAT32   "	// filesystem ID, 8 bytes

	. = start + 0x25A

	.word	0	// reserved
	.word	0	// reserved
	.word	0	// reserved

	. = start + 0x260

1:
	// DS=SS=CS=0, BP=0x7C00, BX=0x7E00
	cld
	//xchgw	%ax, %cx	/* save CX to AX */
	movw	%ds, %cx
	jcxz	1f
2:
	movw	$(error_message - message_func), %bx
message_halt:
	call	message_func
4:
	sti
	hlt
	jmp	4b
1:
	movw	%ss, %cx
	jcxz	1f
	jmp	2b
1:
	movw	%cs, %cx
	jcxz	1f
	jmp	2b
1:
	cmpw	$0x7C00, %bp
	jnz	2b
	cmpw	$0x7E00, %bx
	jnz	2b
	cmpw	%bp, %sp
	ja	2b
	cmpw	$0x7B80, %sp
	jb	2b

	//xchgw	%ax, %cx	/* restore CX from AX */

	/* drive number should be 0x00 or 0x80 */
	movw	0x40(%bp), %dx
	testb	$0x7F, %dl
	jnz	2b

	sti

#if 0
	/* setup BPB: clear hidden sectors */
	xorl	%eax, %eax
	movl	%eax, 0x1C(%bp)
#endif

	movb	$0xFF, %dh	/* partition number for whole drive */

	pushw	%dx

	/* sort out grldr data */

	movw	$(1089 - 1), %cx	/* number of sectors used by grldr */

	movw	$0x800, %bx
	movw	%bx, %ds
	movw	%bx, %es
	movw	$0x0E, %di
	movw	%di, %si
2:
	decw	0x45(%bp)	/* next serial number */
	lodsw			/* get it and will discard it as data */
	pushw	%es
	pushw	%di
	movw	$160, %di
	call	print_hex	# BX changed
	movw	0x45(%bp), %ax
	call	print_hex	# BX changed
	popw	%di
	popw	%es
	movw	$(wrong_SN - message_func), %bx
	cmpw	%ax, 0x45(%bp)	/* check it */
	jne	message_halt	/* wrong serial number, boot failed */

	pushw	%cx
	movw	$(510/2), %cx
	repz movsw
	popw	%cx

	cmpw	$0x8000, %si
	jb	1f
	subw	$0x8000, %si
	movw	%ds, %ax
	addw	$0x800, %ax
	movw	%ax, %ds
1:
	cmpw	$0x8000, %di
	jb	1f
	subw	$0x8000, %di
	movw	%es, %ax
	addw	$0x800, %ax
	movw	%ax, %es
1:
	loop	2b

	/* DX is OK */
	pushw	%dx

	/* clear keyboard buffer */
2:
	movb	$0x01, %ah	/* checkkey */
	int	$0x16

	jz	1f		/* no keypress */

	movb	$0x00, %ah	/* getkey */
	int	$0x16
	jmp	2b
1:
	/* Press P to pause */
	movw	$(pause_message - message_func), %bx
	call	message_func

	//movw	%ss, %ss:0x7FFE	/* 0 means P is not pressed. */
	movl	%ss:0x46c, %eax	/* initial tick */
	addl	$36, %eax	/* wait 2 seconds */
	pushl	%eax
2:
	/* checkkey 'P' */

	movb	$0x01, %ah	/* checkkey */
	int	$0x16

	jz	1f		/* no keypress */

	/* getkey */

	movb	$0x00, %ah	/* getkey */
	int	$0x16


	orb	$0x20, %al
	cmpb	$0x70, %al	/* is "P"? */
	jne	2b		/* no, get next key */
	//je	3f
	//cmpb	$0x70, %ss:0x7FFE	/* P is pressed? */
	//jne	2b		/* no, get next key */
	//jmp	5f
//3:
	/* "P" is pressed. */

	//movb	%al, %ss:0x7FFE	/* P is pressed. */
	popl	%eax
	movl	%ss:0x46c, %eax	/* initial tick */
	addl	$36, %eax	/* wait 2 seconds */
	pushl	%eax
	jmp	2b
1:
	popl	%eax
	pushl	%eax
	movl	%ss:0x46c, %ecx	/* current tick */
	cmpl	%eax, %ecx
	jnb	5f
	subl	$36, %eax
	cmpl	%eax, %ecx
	jnb	2b
	cmpl	$36, %ecx
	jb	2b
5:
	popl	%eax

	popw	%dx
	ljmp	$0, $0x8000
//	pushw	%cs		/* CS=0 */
//	pushw	%bx		/* BX=0x8000 */
//	lret

message_func:
	pushaw
	/* set cursor position at row=15, column=16 */
	xorb	%bh, %bh	/* set page to 0 */
	movw	$0x0F10, %dx	/* DH=row(0 is top), DL=column(0 is left) */
	movb	$0x02, %ah	/* Set-Cursor-Position function call */
	call	int10
	call	3f
3:
	popw	%si
	leaw	(message_func - 3b)(%bx, %si), %si
	call	print_message	/* CS:SI points to message string */
	popaw
	ret

	/* prints string CS:SI (modifies AX BX SI) */
2:
	//xorw	%bx, %bx	/* video page 0 */
	movb	$0x0e, %ah	/* print char in AL */
	call	int10

print_message:

	lodsb	%cs:(%si), %al	/* get token */
	cmpb	$0, %al		/* end of string? */
	jne	2b
	ret
int10:
	pushw	%ds
	pushw	%es
	int	$0x10
	popw	%es
	popw	%ds
	ret

#if 0
#;============================================================================

/* void realmode_printf(const char *format, ...)
 *
 * input:	format is offset in CS segment
 *
 * Usage example:
 *
 * 		pushw	IntegerN
 *		 ... ... ... ...
 * 		pushw	Integer2
 * 		pushw	Integer1
 * 		pushw	$format_string - int13_handler
 *		call	realmode_printf
 * 		addw	$(2*(N+1)), %sp
 *
 * where int13_handle should be the base of the CS segment,
 * and format_string like this:
 *
 * format_string:
 *		 .string "Int1=%x, Int2=%x, ..., IntN=%x\r\n"
 *
 * Currently only %d, %x and %X are implemented.
 */

realmode_printf:
	pushaw
	movw	%sp, %bp
	# bp+18:	format
	# bp+20:	variables
	addw	$18, %bp
	movw	(%bp), %si		# points to format string
	addw	$2, %bp			# (%bp) is the first variable
1:
	cs lodsb
	testb	%al, %al
	jz	1f
	cmpb	$'%', %al
	jne	2f

	#; %d, %u, %x, %X, %4x, %4X
	xorw	%cx, %cx
7:
	cs lodsb
	testb	%al, %al
	jz	1f
	movw	$10, %bx		# base 10
	cmpb	$'d', %al
	jz	4f
	cmpb	$'u', %al
	jz	4f
	cmpb	$'x', %al
	jz	3f
	cmpb	$'X', %al
	jz	3f
	cmpb	$'4', %al
	jne	1b			# unkown directive, continue
	movw	$4, %cx
	jmp	7b
3:
	/* print hexa number */
	movw	$16, %bx		# base 16
4:
	/* print decimal or hexa number */
	pushl	%edi

	movl	$1, %edi
	movw	(%bp), %ax
5:
	xorw	%dx, %dx
	divw	%bx			# AX=quo, DX=rem
	shll	$4, %edi
	orw	%dx, %di
	jcxz	7f
	loop	5b
7:
	testw	%ax, %ax		# end?
	jnz	5b

	/* print the digits in EDI */
	xorw	%bx, %bx	/* video page 0 */
5:
	movw	%di, %ax		# get digit in AL
	shrl	$4, %edi
	jz	7f
	andb	$0x0f, %al
	cmpb	$9, %al
	jbe	6f
	addb	$7, %al			# A, B, C, D, E, F
6:
	addb	$0x30, %al
	movb	$0x0e, %ah	/* print it */
	int	$0x10		/* via TTY mode */
	jmp	5b
7:
	popl	%edi

	addw	$2, %bp			# (%bp) is the next variable
	jmp	1b			# continue
2:
	/* print char in AL */
	xorw	%bx, %bx	/* video page 0 */
	movb	$0x0e, %ah	/* print it */
	int	$0x10		/* via TTY mode */
	jmp	1b			# continue
1:
	popaw
	ret

#;============================================================================
#endif

error_message:

	.ascii	"  It should not get here! Something is wrong!\0"

wrong_SN:

	.ascii	"Wrong Serial Number!\0"

pause_message:

	.ascii	"Repeatedly press P to pause...\0"


	. = start + 0x400

